{{
////////////////////////////////////////////////////////////////////////////////////////////
//                  WM8524.spin
// audio driver for Wolfson Micro WM8524 DAC

// Author: Mark Tillotson
// Updated: 2012-10-31, 2014-01-05
// Designed For: P8X32A
// Version: 1.01

// Provides

PUB  Start (m_clk, b_clk, lr_clk, data, left_chan, right_chan)

     clk pins: m_clk, b_clk, lr_clk,
     data pin: data
     address of long to hold samples to output:  left_chan, right_chan

PUB  Stop

// See end of file for standard MIT licence / terms of use.

// Update History:

// v1.0  - Initial version 2012-10-31
// v1.01 - tidy up 2014-01-05

////////////////////////////////////////////////////////////////////////////////////////////
}}

{{ WM8524_DAC.spin }}

VAR
  long  pinassign
  long  leftptr
  long  rightptr
  long  cog


PUB  Start (m_clk, b_clk, lr_clk, data, left_chan, right_chan)
  pinassign := (data << 24) | (lr_clk << 16) | (b_clk << 8) | (m_clk)
  leftptr  := left_chan
  rightptr := right_chan
  cog := 1 + cognew (@WM8524, @pinassign) 'Launch new cog
  return cog

PUB  Stop
  if cog <> 0
    cogstop (cog-1)
    cog := 0


DAT

{{
   The parameter block consists of a word with the four pin numbers encoded as
   bytes: 
   params   byte     mclkPIN, bclkPIN, lrclkPIN, dinPIN
            long     leftdata
            long     rightdata
}}

              ORG       0                 'Begin at Cog RAM addr 0

WM8524        mov       parm, PAR
              rdlong    temp, parm        ' read the pin specifications
              add       parm, #4
              rdlong    leftaddr, parm
              add       parm, #4
              rdlong    rightaddr, parm

              mov       mclkPIN, temp
              and       mclkPIN, #$3F
              shr       temp, #8
              mov       bclkPIN, temp
              and       bclkPIN, #$3F
              shr       temp, #8
              mov       lrclkPIN, temp
              and       lrclkPIN, #$3F
              shr       temp, #8
              mov       dinPIN, temp
              and       dinPIN, #$3F


              mov       mclkMSK, #1     ' setup output pins (the 3 clock signals)
              shl       mclkMSK, mclkPIN

              mov       bclkMSK, #1
              shl       bclkMSK, bclkPIN

              mov       lrclkMSK, #1
              shl       lrclkMSK, lrclkPIN

              mov       dinMSK, #1
              shl       dinMSK, dinPIN

              or        DIRA, mclkMSK
              or        DIRA, bclkMSK
              or        DIRA, lrclkMSK
              or        DIRA, dinMSK

              mov       FRQA, systemclk         ' setup the mclk frequency
              mov       PHSA, #0

              movs      CTRA, mclkPIN           ' on the right pin
              movd      CTRA, #0                ' no second pin
              movi      CTRA, #$020             ' mode 020

              mov       FRQB, bclk              ' generate the bit clock via counter
              mov       PHSB, #0
              movs      CTRB, bclkPIN
              movd      CTRB, #0
              movi      CTRB, #$020

              mov       time, CNT
              add       time, halfperiod        ' setup LRCLK time accounting

              mov       tout, lrclkMSK
              mov       OUTA, tout              ' LRCLK starts high
              mov       zero, #0

:loop         
              rdlong    left, leftaddr          ' get 24 bit word
              shl       left, #8                ' shift it to top of register
              andn      tout, lrclkMSK          ' set LRCLK low on edge
              mov       n, #24                  ' use Left Justified format, no guard bit
              WAITCNT   time, halfperiod        ' ready for left sample?
              WAITPEQ   zero, bclkMSK           ' sync on previous bclk
              nop                               ' ensure bclk has gone high again before next waitpeq (16 cycles approx)
              nop
              nop
:leftloop     rcl       left, #1  wc            ' shift out one data bit (MSB first)
              muxc      tout, dinMSK            ' put it in right pin position
              WAITPEQ   zero, bclkMSK           ' wait for bclk falling edge
              mov       OUTA, tout              ' drive outputs (LRCLK, DIN, LRCLK)
              nop                               ' ensure bclk has gone high again before next waitpeq (16 cycles approx)
              djnz      n, #:leftloop           ' 32 cycles available per loop, 8 instructions max, here use 6 instrs and waitpeq to synch

              rdlong    right, rightaddr        ' now the right channel, LRCLK is high this time
              shl       right, #8
              or        tout, lrclkMSK          ' set LRCLK high on edge.
              mov       n, #24
              WAITCNT   time, halfperiod        ' wait half an LRclock
              WAITPEQ   zero, bclkMSK           ' wait for the bit clock to be low
              nop
              nop
              nop
:rightloop    rcl       right, #1  wc
              muxc      tout, dinMSK
              WAITPEQ   zero, bclkMSK
              mov       OUTA, tout              ' drive outputs (LRCLK, DIN, LRCLK)
              nop
              djnz      n, #:rightloop

              jmp       #:loop                  ' ready for next sample set



systemclk     long      $20000000  ' 1/8 of system clock, 4 times the bit clock, 256 times lrclk
bclk          long      $08000000
lrclk         long      $00200000
n             long      0
halfperiod    long      $400
zero          long      0

parm          res       1
tout          res       1
left          res       1
right         res       1
leftaddr      res       1
rightaddr     res       1
temp          res       1
time          res       1

mclkPIN       res       1
mclkMSK       res       1
bclkPIN       res       1
bclkMSK       res       1
lrclkPIN      res       1
lrclkMSK      res       1
dinPIN        res       1
dinMSK        res       1

              FIT       $1F0
{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
